"""PyTest plugin for running udf tests."""

import os
import re

import pytest
from google.api_core.exceptions import BadRequest
from google.cloud import bigquery

from bigquery_etl.config import ConfigLoader
from bigquery_etl.util.common import project_dirs

from ..routine.parse_routine import (
    GENERIC_DATASET,
    PROCEDURE_FILE,
    UDF_FILE,
    parse_routines,
)
from .sql_test import dataset

_parsed_routines = None


def parsed_routines():
    """Get cached parsed routines."""
    global _parsed_routines
    if _parsed_routines is None:
        _parsed_routines = {
            routine.filepath: routine
            for project in ConfigLoader.get(
                "routine", "test_projects", fallback=project_dirs()
            )
            for routine in parse_routines(project)
        }

    return _parsed_routines


def pytest_configure(config):
    """Register a custom marker."""
    config.addinivalue_line("markers", "routine: mark routine tests.")


def pytest_collect_file(parent, path):
    """Collect non-python query tests."""
    if "tests/data" not in str(path.dirpath()):
        if path.basename in (UDF_FILE, PROCEDURE_FILE):
            return RoutineFile.from_parent(parent, fspath=path)


class RoutineFile(pytest.File):
    """Routine File."""

    def collect(self):
        """Collect."""
        self.add_marker("routine")
        self.routine = parsed_routines()[self.name]

        for i, query in enumerate(self.routine.tests_full_sql):
            yield RoutineTest.from_parent(
                self, name=f"{self.routine.name}#{i+1}", query=query
            )


class RoutineTest(pytest.Item):
    """Routine Test."""

    def __init__(self, name, parent, query):
        """Initialize."""
        super().__init__(name, parent)
        self.query = query
        if "#xfail" in query:
            self.add_marker(pytest.mark.xfail(strict=True))

    def safe_name(self):
        """Get the name as a valid slug."""
        value = re.sub(r"[^\w\s_]", "", self.name.lower()).strip()
        return re.sub(r"[_\s]+", "_", value)

    def reportinfo(self):
        """Set report title to `self.name`."""
        return super().reportinfo()[:2] + (self.name,)

    def repr_failure(self, excinfo):
        """Skip traceback for api error."""
        if excinfo.errisinstance(BadRequest):
            return str(excinfo.value)
        return super().repr_failure(excinfo)

    def _prunetraceback(self, excinfo):
        """Prune traceback to runtest method."""
        traceback = excinfo.traceback
        ntraceback = traceback.cut(path=__file__)
        excinfo.traceback = ntraceback.filter()

    def runtest(self):
        """Run Test."""
        bq = bigquery.Client()
        dataset_id = self.safe_name()
        if "CIRCLE_BUILD_NUM" in os.environ:
            dataset_id += f"_{os.environ['CIRCLE_BUILD_NUM']}"
        with dataset(bq, dataset_id) as default_dataset:
            job_config = bigquery.QueryJobConfig(
                use_legacy_sql=False, default_dataset=default_dataset
            )
            job = bq.query(
                self.query.replace(GENERIC_DATASET, default_dataset.dataset_id),
                job_config=job_config,
            )
            job.result()



import json
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, Mock, call

import pytest
import smart_open

import bigquery_etl.public_data.publish_gcs_metadata as pgm

TEST_DIR = Path(__file__).parent.parent


class TestPublishGcsMetadata(object):
    test_bucket = "test-bucket"
    project_id = "test-project-id"
    api_version = "v1"
    endpoint = "https://test.endpoint.mozilla.com/"
    sql_dir = TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project"

    mock_blob1 = Mock()
    mock_blob1.name = (
        "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
    )
    mock_blob1.updated = datetime(2020, 4, 3, 11, 30, 1)

    mock_blob2 = Mock()
    mock_blob2.name = "api/v1/tables/test/non_incremental_query/v1/files"
    mock_blob2.updated = datetime(2020, 4, 3, 11, 25, 5)

    mock_storage_client = Mock()
    mock_storage_client.list_blobs.return_value = [mock_blob1, mock_blob2]

    @pytest.fixture(autouse=True)
    def setup(self):
        pass

    def test_dataset_table_version_from_gcs_blob(self):
        mock_blob = Mock()
        mock_blob.name = (
            "api/v1/tables/telemetry_derived/ssl_ratios/v1/files/"
            "2020-04-01/000000001228.json"
        )

        result = pgm.dataset_table_version_from_gcs_blob(mock_blob)

        assert result[0] == "telemetry_derived"
        assert result[1] == "ssl_ratios"
        assert result[2] == "v1"

        mock_blob.name = (
            "api/v1/tables/telemetry_derived/ssl_ratios/v1/files/000000001228.json"
        )

        result = pgm.dataset_table_version_from_gcs_blob(mock_blob)

        assert result[0] == "telemetry_derived"
        assert result[1] == "ssl_ratios"
        assert result[2] == "v1"

        mock_blob.name = "api/v1/tables/telemetry_derived/ssl_ratios/v1/files/"

        result = pgm.dataset_table_version_from_gcs_blob(mock_blob)

        assert result is None

    def test_gcs_table_metadata_no_files(self):
        with pytest.raises(Exception):
            pgm.GcsTableMetadata([], self.endpoint, self.target_dir)

    def test_gcs_table_metadata(self):
        mock_blob = Mock()
        mock_blob.name = (
            "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
        )
        mock_blob.updated = datetime(2020, 4, 3, 11, 30, 1)

        files_path = "api/v1/tables/test/non_incremental_query/v1/files"
        last_updated_path = "api/v1/tables/test/non_incremental_query/v1/last_updated"
        gcs_table_metadata = pgm.GcsTableMetadata(
            [mock_blob], self.endpoint, self.sql_dir
        )

        assert gcs_table_metadata.blobs == [mock_blob]
        assert gcs_table_metadata.endpoint == self.endpoint
        assert gcs_table_metadata.files_path == files_path
        assert gcs_table_metadata.files_uri == self.endpoint + files_path
        assert gcs_table_metadata.dataset == "test"
        assert gcs_table_metadata.table == "non_incremental_query"
        assert gcs_table_metadata.version == "v1"
        assert gcs_table_metadata.metadata.is_incremental() is False
        assert gcs_table_metadata.metadata.is_incremental_export() is False
        assert gcs_table_metadata.metadata.review_bugs() == ["1999999", "12121212"]
        assert gcs_table_metadata.last_updated_path == last_updated_path
        assert gcs_table_metadata.last_updated_uri == self.endpoint + last_updated_path

    def test_gcs_table_metadata_to_json(self):
        mock_blob = Mock()
        mock_blob.name = (
            "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
        )
        mock_blob.updated = datetime(2020, 4, 3, 11, 25, 5)
        files_path = "api/v1/tables/test/non_incremental_query/v1/files"
        last_updated_path = "api/v1/tables/test/non_incremental_query/v1/last_updated"
        gcs_table_metadata = pgm.GcsTableMetadata(
            [mock_blob], self.endpoint, self.sql_dir
        )

        result = gcs_table_metadata.table_metadata_to_json()

        assert len(result.items()) == 7
        assert result["description"] == "Test table for a non-incremental query"
        assert result["friendly_name"] == "Test table for a non-incremental query"
        assert result["incremental"] is False
        assert result["incremental_export"] is False
        review_link = [
            "https://bugzilla.mozilla.org/show_bug.cgi?id=1999999",
            "https://bugzilla.mozilla.org/show_bug.cgi?id=12121212",
        ]
        assert result["review_links"] == review_link
        assert result["files_uri"] == self.endpoint + files_path
        assert result["last_updated"] == self.endpoint + last_updated_path

    def test_gcs_files_metadata_to_json(self):
        mock_blob = Mock()
        mock_blob.name = (
            "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
        )
        mock_blob.updated = datetime(2020, 4, 3, 11, 25, 5)

        json_expected = [
            (
                f"{self.endpoint}"
                "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
            )
        ]

        gcs_table_metadata = pgm.GcsTableMetadata(
            [mock_blob], self.endpoint, self.sql_dir
        )

        result = gcs_table_metadata.files_metadata_to_json()

        assert result == json_expected

    def test_gcs_files_metadata_to_json_incremental(self):
        files = [
            (
                "api/v1/tables/test/incremental_query/v1/files/2020-03-15/"
                "000000000000.json"
            ),
            (
                "api/v1/tables/test/incremental_query/v1/files/2020-03-15/"
                "000000000001.json"
            ),
            (
                "api/v1/tables/test/incremental_query/v1/files/2020-03-16/"
                "000000000000.json"
            ),
        ]

        blobs = []
        for file in files:
            mock_blob = Mock()
            mock_blob.name = file
            mock_blob.updated = datetime(2020, 4, 3, 11, 25, 5)
            blobs.append(mock_blob)

        json_expected = {
            "2020-03-15": [
                (
                    f"{self.endpoint}"
                    "api/v1/tables/test/incremental_query/v1/files/"
                    "2020-03-15/000000000000.json"
                ),
                (
                    f"{self.endpoint}"
                    "api/v1/tables/test/incremental_query/v1/files/"
                    "2020-03-15/000000000001.json"
                ),
            ],
            "2020-03-16": [
                (
                    f"{self.endpoint}"
                    "api/v1/tables/test/incremental_query/v1/files/"
                    "2020-03-16/000000000000.json"
                )
            ],
        }

        gcs_table_metadata = pgm.GcsTableMetadata(blobs, self.endpoint, self.sql_dir)

        result = gcs_table_metadata.files_metadata_to_json()

        assert result == json_expected

    def test_get_public_gcs_table_metadata(self):
        result = list(
            pgm.get_public_gcs_table_metadata(
                self.mock_storage_client,
                self.test_bucket,
                self.api_version,
                self.endpoint,
                self.sql_dir,
            )
        )

        expected = self.endpoint + "api/v1/tables/test/non_incremental_query/v1/files"
        assert len(result) == 1
        assert result[0].files_uri == expected

    def test_publish_all_datasets_metadata(self):
        mock_blob1 = Mock()
        mock_blob1.name = (
            "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
        )
        mock_blob1.updated = datetime(2020, 4, 3, 11, 25, 5)

        mock_blob2 = Mock()
        mock_blob2.name = (
            "api/v1/tables/test/incremental_query/v1/files/2020-03-15/"
            "000000000001.json"
        )
        mock_blob2.updated = datetime(2020, 4, 3, 11, 25, 5)

        gcs_table_metadata = [
            pgm.GcsTableMetadata([mock_blob1], self.endpoint, self.sql_dir),
            pgm.GcsTableMetadata([mock_blob2], self.endpoint, self.sql_dir),
        ]

        mock_out = MagicMock()
        file_handler = MagicMock()
        file_handler.__enter__.return_value = mock_out
        smart_open.open = MagicMock(return_value=file_handler)

        pgm.publish_all_datasets_metadata(gcs_table_metadata, "output.txt")

        all_datasets = TEST_DIR / "data" / "all_datasets.json"
        with open(all_datasets) as f:
            expected_json = json.load(f)
            mock_out.write.assert_called_with(json.dumps(expected_json, indent=4))

    def test_publish_table_metadata(self):
        mock_blob1 = Mock()
        mock_blob1.name = (
            "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
        )
        mock_blob1.updated = datetime(2020, 4, 3, 11, 25, 5)

        mock_blob2 = Mock()
        mock_blob2.name = (
            "api/v1/tables/test/incremental_query/v1/files/2020-03-15/"
            "000000000001.json"
        )
        mock_blob2.updated = datetime(2020, 4, 3, 11, 25, 5)

        gcs_table_metadata = [
            pgm.GcsTableMetadata([mock_blob1], self.endpoint, self.sql_dir),
            pgm.GcsTableMetadata([mock_blob2], self.endpoint, self.sql_dir),
        ]

        mock_out = MagicMock()
        file_handler = MagicMock()
        file_handler.__enter__.return_value = mock_out
        smart_open.open = MagicMock(return_value=file_handler)

        pgm.publish_table_metadata(
            self.mock_storage_client, gcs_table_metadata, self.test_bucket
        )

        metadata_file = TEST_DIR / "data" / "incremental_query_gcs_metadata.json"
        with open(metadata_file) as f:
            expected_incremental_query_json = json.load(f)

        metadata_file = TEST_DIR / "data" / "non_incremental_query_gcs_metadata.json"
        with open(metadata_file) as f:
            expected_non_incremental_query_json = json.load(f)

        mock_out.write.assert_has_calls(
            [
                call(json.dumps(expected_non_incremental_query_json, indent=4)),
                call(json.dumps(expected_incremental_query_json, indent=4)),
            ]
        )

    def test_get_public_gcs_table_metadata_different_projects(self):
        mock_blob1 = Mock()
        mock_blob1.name = (
            "api/v1/tables/test/non_incremental_query/v1/files/000000000000.json"
        )
        mock_blob1.updated = datetime(2020, 4, 3, 11, 30, 1)

        mock_blob2 = Mock()
        mock_blob2.name = "api/v1/tables/test/not_existing/v1/files/000000000000.json"
        mock_blob2.updated = datetime(2020, 4, 3, 11, 25, 5)

        mock_storage_client = Mock()
        mock_storage_client.list_blobs.return_value = [mock_blob1, mock_blob2]

        result = list(
            pgm.get_public_gcs_table_metadata(
                mock_storage_client,
                self.test_bucket,
                self.api_version,
                self.endpoint,
                self.sql_dir,
            )
        )

        expected = self.endpoint + "api/v1/tables/test/non_incremental_query/v1/files"
        assert len(result) == 1
        assert result[0].files_uri == expected



"""Tools for GLAM ETL."""

import os
from pathlib import Path

import rich_click as click
import yaml
from google.cloud import bigquery

from .utils import get_schema, run

ROOT = Path(__file__).parent.parent.parent


def _check_root():
    assert (
        ROOT / "sql" / "moz-fx-data-shared-prod"
    ).exists(), f"{ROOT} is not the project root"


@click.group()
def glam():
    """Tools for GLAM ETL."""
    pass


@glam.group()
def glean():
    """Tools for Glean in GLAM."""
    pass


@glean.command()
@click.option("--project", default="glam-fenix-dev")
@click.option("--dataset", default="glam_etl_dev")
def list_daily(project, dataset):
    """List the start and end dates for clients daily tables."""
    _check_root()
    client = bigquery.Client()
    app_df = client.query(
        rf"""
        WITH
        extracted AS (
            SELECT
                DISTINCT REGEXP_EXTRACT(table_name, "(.*)__") AS app_id,
            FROM
                `{project}`.{dataset}.INFORMATION_SCHEMA.TABLES
            WHERE
                table_name LIKE r"%clients\_daily%" )
        SELECT
            app_id,
            (app_id LIKE r"%\_glam\_%") AS is_logical
        FROM
            extracted
        ORDER BY
            is_logical,
            app_id
    """
    ).to_dataframe()

    query = []
    for row in app_df.itertuples():
        query += [
            f"""
            SELECT
                "{row.app_id}" as app_id,
                {row.is_logical} as is_logical,
                date(min(submission_date)) as earliest,
                date(max(submission_date)) as latest
            FROM
                `{project}`.{dataset}.{row.app_id}__view_clients_daily_scalar_aggregates_v1
            """
        ]

    range_df = (
        client.query("\nUNION ALL\n".join(query))
        .to_dataframe()
        .sort_values(["is_logical", "app_id"])
    )
    click.echo(range_df)


@glean.command()
@click.argument("app-id", type=str)
@click.argument("start-date", type=str)
@click.argument("end-date", type=str)
@click.option("--dataset", type=str, default="glam_etl_dev")
def backfill_daily(app_id, start_date, end_date, dataset):
    """Backfill the daily tables."""
    _check_root()
    run(
        "script/glam/generate_glean_sql",
        cwd=ROOT,
        env={**os.environ, **dict(PRODUCT=app_id, STAGE="daily")},
    )
    run(
        "script/glam/backfill_glean",
        cwd=ROOT,
        env={
            **os.environ,
            **dict(
                DATASET=dataset,
                PRODUCT=app_id,
                STAGE="daily",
                START_DATE=start_date,
                END_DATE=end_date,
                RUN_EXPORT="false",
            ),
        },
    )


@glean.command()
@click.argument("app-id", type=str)
@click.argument("start-date", type=str)
@click.argument("end-date", type=str)
@click.option("--dataset", type=str, default="glam_etl_dev")
def backfill_incremental(app_id, start_date, end_date, dataset):
    """Backfill the incremental tables using existing daily tables.

    To rebuild the table from scratch, drop the clients_scalar_aggregates and
    clients_histogram_aggregates tables.
    """
    _check_root()
    run(
        "script/glam/generate_glean_sql",
        cwd=ROOT,
        env={**os.environ, **dict(PRODUCT=app_id, STAGE="incremental")},
    )
    run(
        "script/glam/backfill_glean",
        cwd=ROOT,
        env={
            **os.environ,
            **dict(
                DATASET=dataset,
                PRODUCT=app_id,
                STAGE="incremental",
                START_DATE=start_date,
                END_DATE=end_date,
                RUN_EXPORT="false",
            ),
        },
    )


@glean.command()
@click.argument("app-id", type=str)
@click.option("--project", default="glam-fenix-dev")
@click.option("--dataset", type=str, default="glam_etl_dev")
@click.option("--bucket", default="glam-fenix-dev-testing")
def export(app_id, project, dataset, bucket):
    """Run the export ETL and write the final csv to a gcs bucket."""
    _check_root()
    run(
        "script/glam/generate_glean_sql",
        cwd=ROOT,
        env={**os.environ, **dict(PRODUCT=app_id, STAGE="incremental")},
    )
    run(
        "script/glam/run_glam_sql",
        cwd=ROOT,
        env={
            **os.environ,
            **dict(
                PROJECT=project,
                DATASET=dataset,
                PRODUCT=app_id,
                STAGE="incremental",
                EXPORT_ONLY="true",
            ),
        },
    )
    run(
        "script/glam/export_csv",
        cwd=ROOT,
        env={
            **os.environ,
            **dict(SRC_PROJECT=project, DATASET=dataset, PRODUCT=app_id, BUCKET=bucket),
        },
    )


@glean.command()
@click.option("--project", default="glam-fenix-dev")
@click.option("--dataset", default="glam_etl")
@click.option("--src-dataset", default="glam_etl_dev")
def update_schemas(project, dataset, src_dataset):
    """Update the schema.yaml files in each query."""
    sql_root = ROOT / "sql" / project / dataset
    for path in sql_root.glob("*"):
        print(f"fetching schema for {path.name}")
        # we can update the schema with the development version of the schema
        schema = dict(fields=get_schema(f"{src_dataset}.{path.name}", project))
        with (path / "schema.yaml").open("w") as fp:
            yaml.dump(schema, fp)


if __name__ == "__main__":
    glam()



#!/usr/bin/env python3
"""clients_daily_scalar_aggregates query generator."""
import argparse
import sys
from typing import Dict, List

from jinja2 import Environment, PackageLoader

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.probe_filters import get_etl_excluded_probes_quickfix

from .utils import get_schema, ping_type_from_table

ATTRIBUTES = ",".join(
    [
        "client_id",
        "ping_type",
        "submission_date",
        "os",
        "app_version",
        "app_build_id",
        "channel",
    ]
)


def render_main(**kwargs):
    """Create a SQL query for the clients_daily_scalar_aggregates dataset."""
    env = Environment(loader=PackageLoader("bigquery_etl", "glam/templates"))
    main_sql = env.get_template("clients_daily_scalar_aggregates_v1.sql")
    return reformat(main_sql.render(**kwargs))


def get_labeled_metrics_sql(probes: Dict[str, List[str]]) -> str:
    """Get the SQL for labeled scalar metrics."""
    probes_struct = []
    for metric_type, _probes in probes.items():
        for probe in _probes:
            probes_struct.append(
                f"('{probe}', '{metric_type}', metrics.{metric_type}.{probe})"
            )

    probes_struct.sort()
    probes_arr = ",\n".join(probes_struct)
    return probes_arr


def get_unlabeled_metrics_sql(probes: Dict[str, List[str]]) -> str:
    """Put together the subsets of SQL required to query scalars or booleans."""
    probe_structs = []
    for probe in probes.pop("boolean", []):
        probe_structs.append(
            (
                f"('{probe}', 'boolean', '', 'false', "
                f"SUM(CAST(NOT metrics.boolean.{probe} AS INT64)))"
            )
        )
        probe_structs.append(
            (
                f"('{probe}', 'boolean', '', 'true', "
                f"SUM(CAST(metrics.boolean.{probe} AS INT64)))"
            )
        )

    for metric_type, _probes in probes.items():
        # timespans are nested within an object that also carries the unit of
        # of time associated with the value
        suffix = ".value" if metric_type == "timespan" else ""
        for probe in _probes:
            for agg_func in ["max", "avg", "min", "sum"]:
                probe_structs.append(
                    (
                        f"('{probe}', '{metric_type}', '', '{agg_func}', "
                        f"{agg_func}(CAST(metrics.{metric_type}.{probe}{suffix} AS NUMERIC)))"
                    )
                )
            probe_structs.append(
                f"('{probe}', '{metric_type}', '', 'count', "
                f"IF(MIN(metrics.{metric_type}.{probe}{suffix}) IS NULL, NULL, COUNT(*)))"
            )

    probe_structs.sort()
    probes_arr = ",\n".join(probe_structs)
    return probes_arr


def get_scalar_metrics(schema: Dict, scalar_type: str) -> Dict[str, List[str]]:
    """Find all scalar probes in a Glean table.

    Metric types are defined in the Glean documentation found here:
    https://mozilla.github.io/glean/book/user/metrics/index.html
    """
    assert scalar_type in ("unlabeled", "labeled")
    metric_type_set = {
        "unlabeled": ["boolean", "counter", "quantity", "timespan"],
        "labeled": ["labeled_counter"],
    }
    scalars: Dict[str, List[str]] = {
        metric_type: [] for metric_type in metric_type_set[scalar_type]
    }
    excluded_metrics = get_etl_excluded_probes_quickfix("fenix")

    # Iterate over every element in the schema under the metrics section and
    # collect a list of metric names.
    for root_field in schema:
        if root_field["name"] != "metrics":
            continue
        for metric_field in root_field["fields"]:
            metric_type = metric_field["name"]
            if metric_type not in metric_type_set[scalar_type]:
                continue
            for field in metric_field["fields"]:
                if field["name"] not in excluded_metrics:
                    scalars[metric_type].append(field["name"])
    return scalars


def main():
    """Print a clients_daily_scalar_aggregates query to stdout."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--no-parameterize",
        action="store_true",
        help="Generate a query without parameters",
    )
    parser.add_argument("--source-table", type=str, help="Name of Glean table")
    parser.add_argument(
        "--product",
        type=str,
        default="org_mozilla_fenix",
    )
    args = parser.parse_args()

    # If set to 1 day, then runs of copy_deduplicate may not be done yet
    submission_date = (
        "date_sub(current_date, interval 2 day)"
        if args.no_parameterize
        else "@submission_date"
    )
    header = (
        "-- Query generated by: python3 -m "
        "bigquery_etl.glam.clients_daily_scalar_aggregates "
        f"--source-table {args.source_table}"
        + (" --no-parameterize" if args.no_parameterize else "")
    )

    schema = get_schema(args.source_table)
    unlabeled_metric_names = get_scalar_metrics(schema, "unlabeled")
    labeled_metric_names = get_scalar_metrics(schema, "labeled")
    unlabeled_metrics = get_unlabeled_metrics_sql(unlabeled_metric_names).strip()
    labeled_metrics = get_labeled_metrics_sql(labeled_metric_names).strip()

    if not unlabeled_metrics and not labeled_metrics:
        print(header)
        print("-- Empty query: no probes found!")
        sys.exit(1)
    print(
        render_main(
            header=header,
            source_table=args.source_table,
            submission_date=submission_date,
            attributes=ATTRIBUTES,
            unlabeled_metrics=unlabeled_metrics,
            labeled_metrics=labeled_metrics,
            ping_type=ping_type_from_table(args.source_table),
        )
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Load Pocket Frontend event data from GCS."""
import uuid
from argparse import ArgumentParser
from datetime import datetime, timedelta

from google.cloud import bigquery

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--date", type=lambda x: datetime.strptime(x, "%Y-%m-%d").date(), required=True
)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--source-bucket", default="moz-fx-data-prod-external-pocket-data")
parser.add_argument("--source-prefix", default="backend_events_for_mozilla")
parser.add_argument("--destination_dataset", default="pocket_derived")
parser.add_argument("--destination_table", default="events_v1")


def main():
    """Load data to temp table and populate destination table via query."""
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    print(f"Running for {args.date}")

    # First, we load the data from GCS to a temp table.
    # The folder naming is done by the Pocket side, so we stick with it.
    uri = f"gs://{args.source_bucket}/{args.source_prefix}/date={args.date}/time=00-00-00-000/*"
    tmp_suffix = uuid.uuid4().hex[0:6]
    tmp_table = f"tmp.pocket_events_{tmp_suffix}"
    result = client.load_table_from_uri(
        uri,
        tmp_table,
        job_config=bigquery.LoadJobConfig(
            write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
            source_format=bigquery.job.SourceFormat.PARQUET,
        ),
    ).result()
    print(f"Loaded {result.output_rows} rows into {tmp_table}")

    # Next, we run a query that populates the destination table partition.
    result = client.query(
        f"""
WITH tmp AS (
SELECT
  * EXCEPT (METADATA, CLIENT_INFO, EVENT_EXTRAS),
  PARSE_JSON(METADATA).metadata AS metadata,
  PARSE_JSON(CLIENT_INFO).client_info AS client_info,
  PARSE_JSON(EVENT_EXTRAS) AS event_extra
FROM {tmp_table} AS t
WHERE
  DATE(SUBMISSION_TIMESTAMP) <= DATE("{args.date.isoformat()}")
)

SELECT
  SUBMISSION_TIMESTAMP as submission_timestamp,
  APP_ID as normalized_app_id,
  STRUCT(
    STRING(client_info.client_id) AS client_id,
    STRING(client_info.user_id) AS user_id,
    STRING(client_info.app_build) AS app_build,
    STRING(client_info.locale) AS locale,
    STRING(client_info.os) AS os,
    STRING(client_info.os_version) AS os_version
  ) AS client_info,
  STRUCT(
    STRUCT(
      STRING(metadata.geo.city) AS city,
      STRING(metadata.geo.country) AS country
    ) AS geo,
    STRUCT(
      STRING(metadata.user_agent.browser) AS browser,
      STRING(metadata.user_agent.os) AS os,
      STRING(metadata.user_agent.version) AS version
    ) AS user_agent
  ) AS metadata,
  EVENT_NAME as event,
  EVENT_TIMESTAMP as event_timestamp,
  event_extra
FROM tmp
        """,
        job_config=bigquery.QueryJobConfig(
            destination=".".join(
                [
                    args.project,
                    args.destination_dataset,
                    f"{args.destination_table}${args.date.strftime('%Y%m%d')}",
                ]
            ),
            write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
        ),
    ).result()
    print(f"Ingested {result.total_rows} rows into {args.destination_table}")

    # Finally, we clean up after ourselves.
    client.delete_table(tmp_table)


if __name__ == "__main__":
    main()



"""Generate documentation for mozfun."""

import itertools
import os
import re
from pathlib import Path

import yaml

from bigquery_etl.config import ConfigLoader
from bigquery_etl.util.mozfun_docs_functions import get_mozfun_parameters

DOCS_FILE = "README.md"
METADATA_FILE = "metadata.yaml"
UDF_FILE = "udf.sql"
PROCEDURE_FILE = "stored_procedure.sql"
SQL_REF_RE = r"@sql\((.+)\)"


def format_url(doc):
    """Create links for urls in documentation."""
    doc = re.sub(r"(?<!\()(https?://[^\s]+)(?!\))", r"<\1>", doc)
    return doc


def load_with_examples(file):
    """Load doc file and replace SQL references with examples."""
    with open(file) as doc_file:
        file_content = doc_file.read()

        path, _ = os.path.split(file)

        for sql_ref in re.findall(SQL_REF_RE, file_content):
            sql_example_file = path / Path(sql_ref)
            with open(sql_example_file) as example_sql:
                md_sql = f"```sql\n{example_sql.read().strip()}\n```"
                file_content = file_content.replace(f"@sql({sql_ref})", md_sql)

    return file_content


def add_source_and_edit(source_url, edit_url):
    """Add links to the function directory and metadata.yaml editor."""
    return f"[Source]({source_url})  |  [Edit]({edit_url})"


def _gen_udf_content(udf_file_path: Path) -> str:
    """Generate markdown documentation content for a udf.sql file."""
    udf_path = udf_file_path.parent
    routine_type = "Stored Procedure" if udf_file_path.name == PROCEDURE_FILE else "UDF"
    udf_content = f"## {udf_path.name} ({routine_type})\n\n"

    metadata, source_link, edit_link = None, None, None
    if (metadata_file := udf_path / "metadata.yaml").exists():
        metadata = yaml.safe_load(metadata_file.read_text())

    if metadata is not None:
        if (description := metadata.get("description")) is not None:
            udf_content += f"{format_url(description)}\n\n"
        source_link = f"{ConfigLoader.get('docs', 'source_url')}/{udf_path}"
        edit_link = f"{ConfigLoader.get('docs', 'edit_url')}/{udf_path}/{METADATA_FILE}"

    if (readme := udf_path / DOCS_FILE).exists():
        udf_content += f"{load_with_examples(readme)}\n\n"

    input_str, output_str = get_mozfun_parameters(udf_file_path.read_text())
    if input_str or output_str:
        udf_content += "\n### Parameters\n\n"
        if input_str:
            udf_content += f"\n**INPUTS**\n\n```\n{input_str}\n```\n\n"
        if output_str:
            udf_content += f"\n**OUTPUTS**\n\n```\n{output_str}\n```\n\n"

    if source_link is not None and edit_link is not None:
        udf_content += f"{add_source_and_edit(source_link, edit_link)}\n\n"

    return udf_content


def generate_udf_docs(out_dir: str, project_dir: str) -> None:
    """Generate documentation for UDFs."""
    project_path = Path(project_dir)
    target_path = Path(out_dir) / project_path.name
    target_path.mkdir(parents=True, exist_ok=True)

    if (project_docs := project_path / DOCS_FILE).exists():
        (target_path / "about.md").write_text(project_docs.read_text())

    # Group by dataset to generate one markdown file per BQ dataset with its UDFs:
    for dataset_path, udf_paths in itertools.groupby(
        sorted(
            list(project_path.glob(f"*/*/{UDF_FILE}"))
            + list(project_path.glob(f"*/*/{PROCEDURE_FILE}"))
        ),
        lambda path: path.parent.parent,  # path is project/dataset/udf/udf.sql
    ):
        if not (udfs_content := (_gen_udf_content(path) for path in udf_paths)):
            continue

        file_content = ""
        if (dataset_docs := dataset_path / DOCS_FILE).exists():
            file_content += f"{dataset_docs.read_text()}\n"
        file_content += "\n".join(udfs_content)
        (target_path / f"{dataset_path.name}.md").write_text(file_content)



"""Tokenize SQL so that it can be formatted."""

import re
import sys
from dataclasses import dataclass, field
from typing import Iterator

# These words get their own line followed by increased indent
TOP_LEVEL_KEYWORDS = [
    # DDL
    "ALTER TABLE IF EXISTS",
    "ALTER TABLE",
    "CLUSTER BY",
    "CREATE OR REPLACE TABLE",
    "CREATE OR REPLACE VIEW",
    "CREATE TABLE IF NOT EXISTS",
    "CREATE VIEW IF NOT EXISTS",
    "CREATE TEMP TABLE",
    "CREATE TABLE",
    "CREATE VIEW",
    "DROP TABLE",
    "DROP VIEW",
    "OPTIONS",
    # DML
    "DELETE FROM",
    "DELETE",
    "INSERT INTO",
    "INSERT",
    "MERGE INTO",
    "MERGE",
    "UPDATE",
    # scripting
    "BEGIN TRANSACTION",
    "BREAK",
    "COMMIT TRANSACTION",
    "COMMIT",
    "CONTINUE",
    "ITERATE",
    "LEAVE",
    "ROLLBACK TRANSACTION",
    "ROLLBACK",
    # SQL
    "AS",  # only when not identified as an AliasSeparator
    "CROSS JOIN",
    "EXCEPT DISTINCT",
    "INTERSECT DISTINCT",
    "FROM",
    "FULL JOIN",
    "FULL OUTER JOIN",
    "GROUP BY",
    "HAVING",
    "INNER JOIN",
    "INTERSECT",
    "JOIN",
    "LEFT JOIN",
    "LEFT OUTER JOIN",
    "LIMIT",
    "ORDER BY",
    "OUTER JOIN",
    "PARTITION BY",
    "QUALIFY",
    "RANGE BETWEEN",
    "RANGE",
    "RIGHT JOIN",
    "RIGHT OUTER JOIN",
    "ROLLUP",
    "ROWS BETWEEN",
    "ROWS",
    "SELECT AS STRUCT",
    "SELECT AS VALUE",
    "SELECT DISTINCT AS STRUCT",
    "SELECT DISTINCT AS VALUE",
    "SELECT DISTINCT",
    "SELECT",
    "UNION ALL",
    "UNION DISTINCT",
    "UNION",
    "VALUES",
    "WHEN MATCHED",
    "WHEN NOT MATCHED BY SOURCE",
    "WHEN NOT MATCHED BY TARGET",
    "WHEN NOT MATCHED",
    "WHERE",
    "WITH(?! OFFSET)",
    "WINDOW",
]
# These words start a new line at the current indent
NEWLINE_KEYWORDS = [
    "ON",
    "USING",
    "WITH OFFSET",
    # UDF
    "CREATE OR REPLACE",
    "CREATE",
    "RETURNS",
    "LANGUAGE",
    # Conditional
    "AND",
    "BETWEEN",
    "ELSE",
    "ELSEIF",
    "END",
    "OR",
    "WHEN",
    "XOR",
]
# These words get capitalized
RESERVED_KEYWORDS = [
    "ALL",
    "AND",
    "ANY",
    "ARRAY",
    "AS",
    "ASC",
    "ASSERT_ROWS_MODIFIED",
    "AT",
    "BETWEEN",
    "BY",
    "CASE",
    "CAST",
    "COLLATE",
    "CONTAINS",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT",
    "DECLARE",
    "DEFAULT",
    "DEFINE",
    "DESC",
    "DISTINCT",
    "ELSE",
    "END",
    "ENUM",
    "ESCAPE",
    "EXCEPT",
    "EXCLUDE",
    "EXISTS",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FOLLOWING",
    "FOR",
    "FROM",
    "FULL",
    "GROUP",
    "GROUPING",
    "GROUPS",
    "HASH",
    "HAVING",
    "IF",
    "IGNORE",
    "IN",
    "INNER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "IS DISTINCT FROM",
    "IS NOT DISTINCT FROM",
    "IS",
    "JOIN",
    "LATERAL",
    "LEFT",
    "LIKE",
    "LIMIT",
    "LOOKUP",
    "MERGE",
    "NATURAL",
    "NEW",
    "NO",
    "NOT",
    "NULL",
    "NULLS",
    "OF",
    "ON",
    "OR",
    "ORDER",
    "OUTER",
    "OVER",
    "PARTITION",
    "PRECEDING",
    "PROTO",
    "RAISE USING MESSAGE",
    "RAISE",
    "RANGE",
    "RECURSIVE",
    "REPLACE",
    "RESPECT",
    "RIGHT",
    "ROLLUP",
    "ROWS",
    "SELECT",
    "SET",
    "SOME",
    "STRUCT",
    "TABLESAMPLE",
    "THEN",
    "TO",
    "TREAT",
    "TRUE",
    "UNBOUNDED",
    "UNION",
    "UNNEST",
    "USING",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
    "WITHIN",
]
# These built-in function names get capitalized
BUILTIN_FUNCTIONS = [
    "ABS",
    "ACOS",
    "ACOSH",
    "AEAD.DECRYPT_BYTES",
    "AEAD.DECRYPT_STRING",
    "AEAD.ENCRYPT",
    "ANY_VALUE",
    "APPROX_COUNT_DISTINCT",
    "APPROX_QUANTILES",
    "APPROX_TOP_COUNT",
    "APPROX_TOP_SUM",
    "ARRAY",
    "ARRAY_AGG",
    "ARRAY_CONCAT",
    "ARRAY_CONCAT_AGG",
    "ARRAY_LENGTH",
    "ARRAY_REVERSE",
    "ARRAY_TO_STRING",
    "ASCII",
    "ASIN",
    "ASINH",
    "ATAN",
    "ATAN2",
    "ATANH",
    "AVG",
    "BIT_AND",
    "BIT_COUNT",
    "BIT_OR",
    "BIT_XOR",
    "BOOL",
    "BYTE_LENGTH",
    "CAST",
    "CBRT",
    "CEIL",
    "CEILING",
    "CHAR_LENGTH",
    "CHARACTER_LENGTH",
    "CHR",
    "COALESCE",
    "CODE_POINTS_TO_BYTES",
    "CODE_POINTS_TO_STRING",
    "COLLATE",
    "CONCAT",
    "CONTAINS_SUBSTR",
    "CORR",
    "COS",
    "COSH",
    "COT",
    "COTH",
    "COUNT",
    "COUNTIF",
    "COVAR_POP",
    "COVAR_SAMP",
    "CSC",
    "CSCH",
    "CUME_DIST",
    "CURRENT_DATE",
    "CURRENT_DATETIME",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "DATE",
    "DATE_ADD",
    "DATE_DIFF",
    "DATE_FROM_UNIX_DATE",
    "DATE_SUB",
    "DATE_TRUNC",
    "DATETIME",
    "DATETIME_ADD",
    "DATETIME_DIFF",
    "DATETIME_SUB",
    "DATETIME_TRUNC",
    "DENSE_RANK",
    "DETERMINISTIC_DECRYPT_BYTES",
    "DETERMINISTIC_DECRYPT_STRING",
    "DETERMINISTIC_ENCRYPT",
    "DIV",
    "ENDS_WITH",
    "ERROR",
    "EXP",
    "EXTERNAL_OBJECT_TRANSFORM",
    "EXTRACT",
    "FARM_FINGERPRINT",
    "FIRST_VALUE",
    "FLOAT64",
    "FLOOR",
    "FORMAT",
    "FORMAT_DATE",
    "FORMAT_DATETIME",
    "FORMAT_TIME",
    "FORMAT_TIMESTAMP",
    "FROM_BASE32",
    "FROM_BASE64",
    "FROM_HEX",
    "GENERATE_ARRAY",
    "GENERATE_DATE_ARRAY",
    "GENERATE_TIMESTAMP_ARRAY",
    "GENERATE_UUID",
    "GREATEST",
    "HLL_COUNT.EXTRACT",
    "HLL_COUNT.INIT",
    "HLL_COUNT.MERGE",
    "HLL_COUNT.MERGE_PARTIAL",
    "IEEE_DIVIDE",
    "IF",
    "IFNULL",
    "INITCAP",
    "INSTR",
    "INT64",
    "IS_INF",
    "IS_NAN",
    "JSON_EXTRACT",
    "JSON_EXTRACT_ARRAY",
    "JSON_EXTRACT_SCALAR",
    "JSON_EXTRACT_STRING_ARRAY",
    "JSON_QUERY",
    "JSON_QUERY_ARRAY",
    "JSON_TYPE",
    "JSON_VALUE",
    "JSON_VALUE_ARRAY",
    "JUSTIFY_DAYS",
    "JUSTIFY_HOURS",
    "JUSTIFY_INTERVAL",
    "KEYS.ADD_KEY_FROM_RAW_BYTES",
    "KEYS.KEYSET_CHAIN",
    "KEYS.KEYSET_FROM_JSON",
    "KEYS.KEYSET_LENGTH",
    "KEYS.KEYSET_TO_JSON",
    "KEYS.NEW_KEYSET",
    "KEYS.NEW_WRAPPED_KEYSET",
    "KEYS.REWRAP_KEYSET",
    "KEYS.ROTATE_KEYSET",
    "KEYS.ROTATE_WRAPPED_KEYSET",
    "LAG",
    "LAST_DAY",
    "LAST_VALUE",
    "LEAD",
    "LEAST",
    "LEFT",
    "LENGTH",
    "LN",
    "LOG",
    "LOG10",
    "LOGICAL_AND",
    "LOGICAL_OR",
    "LOWER",
    "LPAD",
    "LTRIM",
    "MAKE_INTERVAL",
    "MAX",
    "MD5",
    "MIN",
    "MOD",
    "NET.HOST",
    "NET.IP_FROM_STRING",
    "NET.IP_NET_MASK",
    "NET.IP_TO_STRING",
    "NET.IP_TRUNC",
    "NET.IPV4_FROM_INT64",
    "NET.IPV4_TO_INT64",
    "NET.PUBLIC_SUFFIX",
    "NET.REG_DOMAIN",
    "NET.SAFE_IP_FROM_STRING",
    "NORMALIZE",
    "NORMALIZE_AND_CASEFOLD",
    "NTH_VALUE",
    "NTILE",
    "NULLIF",
    "OCTET_LENGTH",
    "OFFSET",
    "ORDINAL",
    "PARSE_BIGNUMERIC",
    "PARSE_DATE",
    "PARSE_DATETIME",
    "PARSE_JSON",
    "PARSE_NUMERIC",
    "PARSE_TIME",
    "PARSE_TIMESTAMP",
    "PERCENT_RANK",
    "PERCENTILE_CONT",
    "PERCENTILE_DISC",
    "POW",
    "POWER",
    "RAND",
    "RANGE_BUCKET",
    "RANK",
    "REGEXP_CONTAINS",
    "REGEXP_EXTRACT",
    "REGEXP_EXTRACT_ALL",
    "REGEXP_INSTR",
    "REGEXP_REPLACE",
    "REGEXP_SUBSTR",
    "REPEAT",
    "REPLACE",
    "REVERSE",
    "RIGHT",
    "ROUND",
    "ROW_NUMBER",
    "RPAD",
    "RTRIM",
    "S2_CELLIDFROMPOINT",
    "S2_COVERINGCELLIDS",
    "SAFE_ADD",
    "SAFE_CAST",
    "SAFE_CONVERT_BYTES_TO_STRING",
    "SAFE_DIVIDE",
    "SAFE_MULTIPLY",
    "SAFE_NEGATE",
    "SAFE_OFFSET",
    "SAFE_ORDINAL",
    "SAFE_SUBTRACT",
    "SEC",
    "SECH",
    "SESSION_USER",
    "SHA1",
    "SHA256",
    "SHA512",
    "SIGN",
    "SIN",
    "SINH",
    "SOUNDEX",
    "SPLIT",
    "SQRT",
    "ST_ANGLE",
    "ST_AREA",
    "ST_ASBINARY",
    "ST_ASGEOJSON",
    "ST_ASTEXT",
    "ST_AZIMUTH",
    "ST_BOUNDARY",
    "ST_BOUNDINGBOX",
    "ST_BUFFER",
    "ST_BUFFERWITHTOLERANCE",
    "ST_CENTROID",
    "ST_CENTROID_AGG",
    "ST_CLOSESTPOINT",
    "ST_CLUSTERDBSCAN",
    "ST_CONTAINS",
    "ST_CONVEXHULL",
    "ST_COVEREDBY",
    "ST_COVERS",
    "ST_DIFFERENCE",
    "ST_DIMENSION",
    "ST_DISJOINT",
    "ST_DISTANCE",
    "ST_DUMP",
    "ST_DWITHIN",
    "ST_ENDPOINT",
    "ST_EQUALS",
    "ST_EXTENT",
    "ST_EXTERIORRING",
    "ST_GEOGFROM",
    "ST_GEOGFROMGEOJSON",
    "ST_GEOGFROMTEXT",
    "ST_GEOGFROMWKB",
    "ST_GEOGPOINT",
    "ST_GEOGPOINTFROMGEOHASH",
    "ST_GEOHASH",
    "ST_GEOMETRYTYPE",
    "ST_INTERIORRINGS",
    "ST_INTERSECTION",
    "ST_INTERSECTS",
    "ST_INTERSECTSBOX",
    "ST_ISCLOSED",
    "ST_ISCOLLECTION",
    "ST_ISEMPTY",
    "ST_ISRING",
    "ST_LENGTH",
    "ST_MAKELINE",
    "ST_MAKEPOLYGON",
    "ST_MAKEPOLYGONORIENTED",
    "ST_MAXDISTANCE",
    "ST_NPOINTS",
    "ST_NUMGEOMETRIES",
    "ST_NUMPOINTS",
    "ST_PERIMETER",
    "ST_POINTN",
    "ST_SIMPLIFY",
    "ST_SNAPTOGRID",
    "ST_STARTPOINT",
    "ST_TOUCHES",
    "ST_UNION",
    "ST_UNION_AGG",
    "ST_WITHIN",
    "ST_X",
    "ST_Y",
    "STARTS_WITH",
    "STDDEV",
    "STDDEV_POP",
    "STDDEV_SAMP",
    "STRING",
    "STRING_AGG",
    "STRPOS",
    "SUBSTR",
    "SUBSTRING",
    "SUM",
    "TAN",
    "TANH",
    "TIME",
    "TIME_ADD",
    "TIME_DIFF",
    "TIME_SUB",
    "TIME_TRUNC",
    "TIMESTAMP",
    "TIMESTAMP_ADD",
    "TIMESTAMP_DIFF",
    "TIMESTAMP_MICROS",
    "TIMESTAMP_MILLIS",
    "TIMESTAMP_SECONDS",
    "TIMESTAMP_SUB",
    "TIMESTAMP_TRUNC",
    "TO_BASE32",
    "TO_BASE64",
    "TO_CODE_POINTS",
    "TO_HEX",
    "TO_JSON",
    "TO_JSON_STRING",
    "TRANSLATE",
    "TRIM",
    "TRUNC",
    "UNICODE",
    "UNIX_DATE",
    "UNIX_MICROS",
    "UNIX_MILLIS",
    "UNIX_SECONDS",
    "UPPER",
    "VAR_POP",
    "VAR_SAMP",
    "VARIANCE",
]
QUOTE = "\"\"\"|'''|\"|'"
# strings contain any character, with backslash always followed by one more character
STRING_CONTENT = r"\\.|[^\\]"


def _keyword_pattern(words):
    """
    Compile a regex that will match the given list of keywords.

    Match one or more whitespace characters between words patterns, and only
    match whole words.
    """
    return re.compile(
        "(?:" + "|".join(pattern.replace(" ", r"\s+") for pattern in words) + r")\b",
        re.IGNORECASE,
    )


@dataclass
class Token:
    """Abstract token class."""

    value: str
    pattern: re.Pattern = field(init=False, repr=False)

    def __post_init__(self):
        """Enable post-init for child classes."""
        pass


class Comment(Token):
    """Comment abstract class.

    Comments must match preceding whitespace, but not match multiple preceding
    lines or trailing whitespace.
    """

    _format_off = re.compile(r"\bformat\s*:?\s*off\b")
    _format_on = re.compile(r"\bformat\s*:?\s*on\b")

    def __post_init__(self):
        """Detect format off/on comments."""
        self.format_off = self._format_off.search(self.value) is not None
        self.format_on = self._format_on.search(self.value) is not None


class LineComment(Comment):
    """Comment that spans to the end of the current line, sans trailing whitespace."""

    pattern = re.compile(r"\n?[^\S\n]*(#|--)([^\n]*\S)?")


class BlockComment(Comment):
    """Comment that may span multiple lines."""

    pattern = re.compile(r"\n?[^\S\n]*/\*.*?\*/", re.DOTALL)


class Whitespace(Token):
    """One or more whitespace characters on a single line."""

    pattern = re.compile(r"\s[^\S\n]*")


class ReservedKeyword(Token):
    """Token that gets capitalized and separates words with a single space."""

    pattern = _keyword_pattern(RESERVED_KEYWORDS)


class SpaceBeforeBracketKeyword(ReservedKeyword):
    """Keyword that should be separated by a space from a following opening bracket."""

    pattern = _keyword_pattern(
        [
            "IN",
            r"\* EXCEPT",
            r"\* REPLACE",
            "NOT",
            "OVER",
            "IS DISTINCT FROM",
            "IS NOT DISTINCT FROM",
        ]
    )


class BlockKeyword(ReservedKeyword):
    """Keyword that separates indented blocks, such as conditionals."""


class BlockStartKeyword(BlockKeyword):
    """Keyword that gets its own line followed by increased indent."""

    pattern = _keyword_pattern(
        [
            "CREATE OR REPLACE PROCEDURE",
            "CREATE PROCEDURE IF NOT EXISTS",
            "CREATE PROCEDURE",
            # negative lookahead prevents matching IF function
            r"IF(?!(\s|\n)*[(])",
            "WHILE",
            "LOOP",
            "CASE",
        ]
    )


class BlockEndKeyword(BlockKeyword):
    """Keyword that gets its own line preceded by decreased indent."""

    pattern = _keyword_pattern(["END( (WHILE|LOOP|IF))?"])


class BlockMiddleKeyword(BlockStartKeyword, BlockEndKeyword):
    """Keyword that ends one indented block and starts another."""

    pattern = _keyword_pattern(
        [r"BEGIN(?!\s+TRANSACTION\b)", "EXCEPTION WHEN ERROR THEN", "ELSEIF", "DO"]
    )


class AliasSeparator(SpaceBeforeBracketKeyword):
    """Keyword separating an expression from an alias.

    May be followed by an alias identifier that would otherwise be a reserved keyword.

    Must not be followed by the keyword WITH, SELECT, STRUCT or ARRAY.
    """

    pattern = re.compile(
        r"AS(?=\s+(?!(WITH|SELECT|STRUCT|ARRAY)\b)[a-z_`({])", re.IGNORECASE
    )


class NewlineKeyword(SpaceBeforeBracketKeyword):
    """Keyword that should start a new line."""

    pattern = _keyword_pattern(NEWLINE_KEYWORDS)


class TopLevelKeyword(NewlineKeyword):
    """Keyword that should get its own line followed by increased indent."""

    pattern = _keyword_pattern(TOP_LEVEL_KEYWORDS)


class CaseSubclause(NewlineKeyword):
    """Subclause within a CASE."""

    pattern = _keyword_pattern(["WHEN"])


class MaybeCaseSubclause(ReservedKeyword):
    """Keyword that needs context to determine whether it is for a CASE or an IF."""

    pattern = _keyword_pattern(["THEN", "ELSE"])


class AngleBracketKeyword(ReservedKeyword):
    """Keyword indicating that if the next token is '<' it is a bracket."""

    pattern = _keyword_pattern(["ARRAY", "STRUCT"])


class Identifier(Token):
    """Identifier for a column, table, or other database object."""

    pattern = re.compile(r"[A-Za-z_][A-Za-z_0-9]*|`(?:\\.|[^\\`])+`")


class QualifiedIdentifier(Identifier):
    """Fully or partially qualified identifier for a column, table, or other database object."""

    pattern = re.compile(
        rf"(?:(?:{Identifier.pattern.pattern})\.)+(?:{Identifier.pattern.pattern})"
    )


class BuiltInFunctionIdentifier(Identifier):
    """Identifier for a built-in function."""

    pattern = re.compile(
        r"(?:SAFE\.)?(?:"
        + "|".join(re.escape(f) for f in BUILTIN_FUNCTIONS)
        + r")(?=\()",
        re.IGNORECASE,
    )


class QueryParameter(Identifier):
    """Query parameter."""

    pattern = re.compile("@[A-Za-z_][A-Za-z_0-9]*")


class Literal(Token):
    """A constant value of a built-in data type."""

    pattern = re.compile(
        # String literal
        rf"(?:r?b|b?r)?({QUOTE})(?:{STRING_CONTENT})*?\1"
        # Hexadecimal integer literal
        "|0[xX][0-9a-fA-F]+"
        # Decimal integer or float literal
        r"|\d+\.?\d*(?:[Ee][+-]?)?\d*"
    )


class JinjaExpression(Token):
    """Jinja expression delimiters {{ }}."""

    pattern = re.compile(r"{{.*?}}", re.DOTALL)


class JinjaStatement(Token):
    """Jinja statement delimiters {% %}.

    May be followed by no whitespace or a new line and increased indent.
    """

    pattern = re.compile(r"{%.*?%}", re.DOTALL)


class JinjaBlockStatement(JinjaStatement):
    """Statements that start and/or end Jinja blocks."""


class JinjaBlockStart(JinjaBlockStatement):
    """Jinja block starts get their own line followed by increased indent."""

    pattern = re.compile(r"{%[-+]? *(block|call|filter|for|if|macro)\b.*?%}", re.DOTALL)


class JinjaBlockEnd(JinjaBlockStatement):
    """Jinja block ends get their own line preceded by decreased indent."""

    pattern = re.compile(
        r"{%[-+]? *end(block|call|filter|for|if|macro)\b.*?%}", re.DOTALL
    )


class JinjaBlockMiddle(JinjaBlockEnd, JinjaBlockStart):
    """Ends one indented Jinja block and starts another."""

    pattern = re.compile(r"{%[-+]? *(elif|else)\b.*?%}", re.DOTALL)


class JinjaComment(BlockComment):
    """Jinja comment that may span multiple lines."""

    pattern = re.compile(r"\n?[^\S\n]*{#.*?#}", re.DOTALL)


class OpeningBracket(Token):
    """Opening bracket or parenthesis.

    May be followed by no whitespace or a new line and increased indent.
    """

    pattern = re.compile(r"[([]")


class ClosingBracket(Token):
    """Closing bracket or parenthesis.

    May be preceded by a new line and decreased indent, or no whitespace.
    """

    pattern = re.compile(r"[\])]")


class MaybeOpeningAngleBracket(Token):
    """Token that needs context to determine whether it is an operator or bracket."""

    pattern = re.compile(r"<")


class MaybeClosingAngleBracket(Token):
    """Token that needs context to determine whether it is an operator or bracket."""

    pattern = re.compile(r">")


class ExpressionSeparator(Token):
    """Token that separates expressions.

    Should not be preceded by whitespace and may be followed by a space or a
    new line.
    """

    pattern = re.compile(r",")


class StatementSeparator(Token):
    """Token that separates statements.

    Should not be preceded by whitespace and should be followed by a new line
    and reset indent.
    """

    pattern = re.compile(r";")


class Operator(Token):
    """Operator."""

    pattern = re.compile(r"<<|>>|>=|<=|=>|<>|!=|.")


class ConcatenationOperator(Token):
    """Concatenation operator."""

    pattern = re.compile(r"\|\|")


class FieldAccessOperator(Operator):
    """Operator for field access.

    May use whitespace different from other operators.

    May be followed by an identifier that would otherwise be a reserved
    keyword.
    """

    pattern = re.compile(r"\.")


BIGQUERY_TOKEN_PRIORITY = [
    LineComment,
    BlockComment,
    JinjaComment,
    Whitespace,
    JinjaExpression,
    JinjaBlockStart,
    JinjaBlockMiddle,
    JinjaBlockEnd,
    JinjaStatement,
    BlockMiddleKeyword,
    BlockStartKeyword,
    BlockEndKeyword,
    AliasSeparator,
    TopLevelKeyword,
    MaybeCaseSubclause,
    CaseSubclause,
    NewlineKeyword,
    AngleBracketKeyword,
    SpaceBeforeBracketKeyword,
    ReservedKeyword,
    ConcatenationOperator,
    Literal,
    BuiltInFunctionIdentifier,
    QualifiedIdentifier,
    Identifier,
    QueryParameter,
    OpeningBracket,
    ClosingBracket,
    MaybeOpeningAngleBracket,
    MaybeClosingAngleBracket,
    FieldAccessOperator,
    ExpressionSeparator,
    StatementSeparator,
    Operator,
]


def tokenize(query, token_priority=BIGQUERY_TOKEN_PRIORITY) -> Iterator[Token]:
    """Split query into a series of tokens."""
    open_blocks: list[BlockStartKeyword] = []
    open_angle_brackets = 0
    angle_bracket_is_operator = True
    reserved_keyword_is_identifier = False
    while query:
        for token_type in token_priority:
            match = token_type.pattern.match(query)
            if not match:
                continue
            token = token_type(match.group())
            # handle stateful matches
            if isinstance(token, MaybeCaseSubclause):
                if open_blocks and open_blocks[-1].value.upper() == "CASE":
                    token = CaseSubclause(token.value)
                else:
                    token = BlockMiddleKeyword(token.value)
            elif isinstance(token, MaybeOpeningAngleBracket):
                if angle_bracket_is_operator:
                    continue  # prevent matching operator as opening bracket
                token = OpeningBracket(token.value)
                open_angle_brackets += 1
            elif isinstance(token, MaybeClosingAngleBracket):
                if angle_bracket_is_operator:
                    continue  # prevent matching operator as closing bracket
                token = ClosingBracket(token.value)
                open_angle_brackets -= 1
            elif (
                reserved_keyword_is_identifier
                and isinstance(token, ReservedKeyword)
                and Identifier.pattern.match(token.value) is not None
            ):
                continue  # prevent matching identifier as keyword
            yield token
            length = len(token.value)
            query = query[length:]
            # update stateful conditions for next token
            if isinstance(token, BlockEndKeyword) and open_blocks:
                open_blocks.pop()
            if isinstance(token, BlockStartKeyword):
                open_blocks.append(token)
            if not isinstance(token, (Comment, Whitespace)):
                # angle brackets are operators unless already in angle bracket
                # block or preceded by an AngleBracketKeyword
                angle_bracket_is_operator = not (
                    open_angle_brackets > 0 or isinstance(token, AngleBracketKeyword)
                )
                # field access operator may be followed by an identifier that
                # would otherwise be a reserved keyword.
                reserved_keyword_is_identifier = isinstance(
                    token, (FieldAccessOperator, AliasSeparator)
                )
            break
        else:
            raise ValueError(f"Could not determine next token in {query!r}")


if __name__ == "__main__":
    # entrypoint for inspecting tokenize results
    for token in tokenize(sys.stdin.read()):
        print(f"{type(token).__name__}: {token.value!r}")



#!/usr/bin/env python3

"""Search for tables and user ids that may be eligible for self serve deletion."""

import re
import warnings
from argparse import ArgumentParser

from google.cloud import bigquery

from ..util import standard_args
from .config import SEARCH_IGNORE_FIELDS, SEARCH_IGNORE_TABLES, SHARED_PROD

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "-p",
    "--project",
    "--project_id",
    "--project-id",
    default=SHARED_PROD,
    help=f"ID of the project in which to find tables; defaults to {SHARED_PROD}",
)
standard_args.add_log_level(parser, default=None)
standard_args.add_table_filter(parser)

ID_PATTERN = re.compile(r"(\b|_)id")
IGNORE_PATTERN = re.compile(
    "|".join(
        [
            "activation_id",
            "addon_id",
            "application_id",
            "batch_id",
            "bucket_id",
            "bug_id",
            "build_id",
            "campaign_id",
            "changeset_id",
            "crash_id",
            "device_id",
            "distribution_id",
            "document_id",
            "error_id",
            "experiment_id",
            "extension_id",
            "encryption_key_id",
            "insert_id",
            "message_id",
            "model_id",
            "network_id",
            "page_id",
            "partner_id",
            "product_id",
            "run_id",
            "setter_id",
            "survey_id",
            "sample_id",
            "session_id",
            "subsys(tem)?_id",
            "thread_id",
            "tile_id",
            "vendor_id",
            "id_bucket",
            r"active_experiment\.id",
            r"theme\.id",
            r"tiles\[]\.id",
            r"spoc_fills\[]\.id",
            r"devices\[]\.id",
            r"application\.id",
            r"environment\.id",
        ]
    )
)
DATASET_PATTERN = re.compile(".*_(stable|derived)")


def find_id_fields(fields, prefix=""):
    """Recursively locate potential ids in fields."""
    for field in fields:
        name = prefix + field.name
        if field.field_type == "RECORD":
            prefix += field.name + ("[]" if field.mode == "REPEATED" else "") + "."
            yield from find_id_fields(field.fields, prefix)
        elif ID_PATTERN.search(name) and not IGNORE_PATTERN.search(name):
            yield name


def find_target_tables(project, table_filter):
    """Search for potential new ids and new tables with ids."""
    client = bigquery.Client()
    for dataset in client.list_datasets(project):
        if not DATASET_PATTERN.match(dataset.dataset_id):
            continue
        for table_ref in client.list_tables(dataset.reference):
            if table_ref.table_type != "TABLE":
                continue
            table = f"{table_ref.dataset_id}.{table_ref.table_id}"
            if not table_filter(table) or table in SEARCH_IGNORE_TABLES:
                continue
            for field in find_id_fields(client.get_table(table_ref).schema):
                result = table, field
                if result not in SEARCH_IGNORE_FIELDS:
                    yield result


def main():
    """Print results of find_target_tables."""
    args = parser.parse_args()
    for table, field in find_target_tables(args.project, args.table_filter):
        print(f"table={table!r}, field={field!r}")


if __name__ == "__main__":
    warnings.filterwarnings("ignore", module="google.auth._default")
    main()



"""Utility functions used by backfills."""

import re
import sys
from datetime import date
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click
from google.cloud import bigquery
from google.cloud.exceptions import NotFound

from bigquery_etl.backfill.parse import BACKFILL_FILE, Backfill, BackfillStatus
from bigquery_etl.metadata.parse_metadata import (
    DATASET_METADATA_FILE,
    METADATA_FILE,
    DatasetMetadata,
    Metadata,
)
from bigquery_etl.util import extract_from_query_path

QUALIFIED_TABLE_NAME_RE = re.compile(
    r"(?P<project_id>[a-zA-z0-9_-]+)\.(?P<dataset_id>[a-zA-z0-9_-]+)\.(?P<table_id>[a-zA-Z0-9-_$]+)"
)

BACKFILL_DESTINATION_PROJECT = "moz-fx-data-shared-prod"
BACKFILL_DESTINATION_DATASET = "backfills_staging_derived"

# currently only supporting backfilling tables with workgroup access: mozilla-confidential.
VALID_WORKGROUP_MEMBER = ["workgroup:mozilla-confidential"]


def get_entries_from_qualified_table_name(
    sql_dir, qualified_table_name, status=None
) -> List[Backfill]:
    """Return backfill entries from qualified table name."""
    backfills = []

    project, dataset, table = qualified_table_name_matching(qualified_table_name)
    table_path = Path(sql_dir) / project / dataset / table

    if not table_path.exists():
        click.echo(f"{project}.{dataset}.{table}" + " does not exist")
        sys.exit(1)

    backfill_file = get_backfill_file_from_qualified_table_name(
        sql_dir, qualified_table_name
    )

    if backfill_file.exists():
        backfills = Backfill.entries_from_file(backfill_file, status)

    return backfills


def get_qualified_table_name_to_entries_map_by_project(
    sql_dir, project_id: str, status: Optional[str] = None
) -> Dict[str, List[Backfill]]:
    """Return backfill entries from project or all projects if project_id=None is given."""
    backfills_dict: dict = {}

    project_id_glob = project_id if project_id is not None else "*"

    backfill_files = Path(sql_dir).glob(f"{project_id_glob}/*/*/{BACKFILL_FILE}")
    for backfill_file in backfill_files:
        project, dataset, table = extract_from_query_path(backfill_file)
        qualified_table_name = f"{project}.{dataset}.{table}"

        entries = get_entries_from_qualified_table_name(
            sql_dir, qualified_table_name, status
        )

        if entries:
            backfills_dict[qualified_table_name] = entries

    return backfills_dict


def get_backfill_file_from_qualified_table_name(sql_dir, qualified_table_name) -> Path:
    """Return backfill file from qualified table name."""
    project, dataset, table = qualified_table_name_matching(qualified_table_name)
    path = Path(sql_dir)
    query_path = path / project / dataset / table
    backfill_file = query_path / BACKFILL_FILE

    return backfill_file


# TODO: It would be better to take in a backfill object.
def get_backfill_staging_qualified_table_name(qualified_table_name, entry_date) -> str:
    """Return full table name where processed backfills are stored."""
    _, dataset, table = qualified_table_name_matching(qualified_table_name)
    backfill_table_id = f"{dataset}__{table}_{entry_date}".replace("-", "_")

    return f"{BACKFILL_DESTINATION_PROJECT}.{BACKFILL_DESTINATION_DATASET}.{backfill_table_id}"


def get_backfill_backup_table_name(qualified_table_name: str, entry_date: date) -> str:
    """Return full table name where backup of production table is stored."""
    _, dataset, table = qualified_table_name_matching(qualified_table_name)
    cloned_table_id = f"{dataset}__{table}_backup_{entry_date}".replace("-", "_")

    return f"{BACKFILL_DESTINATION_PROJECT}.{BACKFILL_DESTINATION_DATASET}.{cloned_table_id}"


def validate_depends_on_past(sql_dir, qualified_table_name) -> bool:
    """
    Check if the table depends on past.

    Managed backfills currently do not support tables that depends on past.
    """
    project, dataset, table = qualified_table_name_matching(qualified_table_name)
    table_metadata_path = Path(sql_dir) / project / dataset / table / METADATA_FILE

    table_metadata = Metadata.from_file(table_metadata_path)

    if "depends_on_past" in table_metadata.scheduling:
        return not table_metadata.scheduling[
            "depends_on_past"
        ]  # skip if depends_on_past

    return True


def validate_metadata_workgroups(sql_dir, qualified_table_name) -> bool:
    """
    Check if either table or dataset metadata workgroup is valid.

    The backfill staging dataset currently only support backfilling datasets and tables for workgroup:mozilla-confidential.
    """
    project, dataset, table = qualified_table_name_matching(qualified_table_name)
    query_file = Path(sql_dir) / project / dataset / table / "query.sql"
    dataset_path = Path(sql_dir) / project / dataset
    dataset_metadata_path = dataset_path / DATASET_METADATA_FILE
    table_metadata_path = dataset_path / table / METADATA_FILE

    if not query_file.exists():
        click.echo("No query.sql file found for {}", qualified_table_name)
        sys.exit(1)

    # check dataset level metadata
    try:
        dataset_metadata = DatasetMetadata.from_file(dataset_metadata_path)
        dataset_workgroup_access = dataset_metadata.workgroup_access
        dataset_default_table_workgroup_access = (
            dataset_metadata.default_table_workgroup_access
        )

    except FileNotFoundError as e:
        raise ValueError(
            f"Unable to validate workgroups for {qualified_table_name} in dataset metadata file."
        ) from e

    if _validate_workgroup_members(dataset_workgroup_access, DATASET_METADATA_FILE):
        return True

    # check table level metadata
    try:
        table_metadata = Metadata.from_file(table_metadata_path)
        table_workgroup_access = table_metadata.workgroup_access

        if _validate_workgroup_members(table_workgroup_access, METADATA_FILE):
            return True

    except FileNotFoundError:
        # default table workgroup access is applied to table if metadata.yaml file is missing
        if _validate_workgroup_members(
            dataset_default_table_workgroup_access, DATASET_METADATA_FILE
        ):
            return True

    return False


def _validate_workgroup_members(workgroup_access, metadata_filename):
    """Return True if workgroup members is valid (workgroup:mozilla-confidential)."""
    if workgroup_access:
        for workgroup in workgroup_access:
            if metadata_filename == METADATA_FILE:
                members = workgroup.members
            elif metadata_filename == DATASET_METADATA_FILE:
                members = workgroup["members"]

            if members == VALID_WORKGROUP_MEMBER:
                return True

    return False


def qualified_table_name_matching(qualified_table_name) -> Tuple[str, str, str]:
    """Match qualified table name pattern."""
    if match := QUALIFIED_TABLE_NAME_RE.match(qualified_table_name):
        project_id = match.group("project_id")
        dataset_id = match.group("dataset_id")
        table_id = match.group("table_id")
    else:
        raise AttributeError(
            "Qualified table name must be named like:" + " <project>.<dataset>.<table>"
        )

    return project_id, dataset_id, table_id


def get_scheduled_backfills(
    sql_dir,
    project: str,
    qualified_table_name: Optional[str] = None,
    status: Optional[str] = None,
) -> Dict[str, Backfill]:
    """Return backfill entries to initiate or complete."""
    client = bigquery.Client(project=project)

    if qualified_table_name:
        backfills_dict = {
            qualified_table_name: get_entries_from_qualified_table_name(
                sql_dir, qualified_table_name, status
            )
        }
    else:
        backfills_dict = get_qualified_table_name_to_entries_map_by_project(
            sql_dir, project, status
        )

    backfills_to_process_dict = {}

    for qualified_table_name, entries in backfills_dict.items():
        # do not return backfill if depends on past
        if not validate_depends_on_past(sql_dir, qualified_table_name):
            continue

        # do not return backfill if not mozilla-confidential
        if not validate_metadata_workgroups(sql_dir, qualified_table_name):
            continue

        if not entries:
            continue

        if BackfillStatus.INITIATE.value == status and (len(entries)) > 1:
            raise ValueError(
                f"More than one backfill entry for {qualified_table_name} found with status: {status}."
            )

        entry_to_process = entries[0]

        if (
            BackfillStatus.INITIATE.value == status
            and _should_initiate(client, entry_to_process, qualified_table_name)
        ) or (
            BackfillStatus.COMPLETE.value == status
            and _should_complete(client, entry_to_process, qualified_table_name)
        ):
            backfills_to_process_dict[qualified_table_name] = entry_to_process

    return backfills_to_process_dict


def _should_initiate(
    client: bigquery.Client, backfill: Backfill, qualified_table_name: str
) -> bool:
    """Determine whether a backfill should be initiated.

    Return true if the backfill is in Initiate status and the staging data does not yet exist.
    """
    if backfill.status != BackfillStatus.INITIATE:
        return False

    staging_table = get_backfill_staging_qualified_table_name(
        qualified_table_name, backfill.entry_date
    )

    if _table_exists(client, staging_table):
        return False

    return True


def _should_complete(
    client: bigquery.Client, backfill: Backfill, qualified_table_name: str
) -> bool:
    """Determine whether a backfill should be completed.

    Return true if the backfill is in Complete status, the staging data exists, and the backup does not yet exist.
    """
    if backfill.status != BackfillStatus.COMPLETE:
        return False

    staging_table = get_backfill_staging_qualified_table_name(
        qualified_table_name, backfill.entry_date
    )
    if not _table_exists(client, staging_table):
        click.echo(f"Backfill staging table does not exist: {staging_table}")
        return False

    backup_table_name = get_backfill_backup_table_name(
        qualified_table_name, backfill.entry_date
    )
    if _table_exists(client, backup_table_name):
        click.echo(f"Backfill backup table already exists: {backup_table_name}")
        return False

    return True


def _table_exists(client: bigquery.Client, qualified_table_name: str) -> bool:
    try:
        client.get_table(qualified_table_name)
        return True
    except NotFound:
        return False



"""Testing data for query."""

from pathlib import Path

import yaml

ROOT = Path(__file__).parent

SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"

SCALAR_BUCKET_COUNTS = [
    {
        "agg_type": "histogram",
        "app_build_id": "*",
        "app_version": 84,
        "bucket": "2.00",
        "bucket_count": 3,
        "channel": "*",
        "client_agg_type": "count",
        "count": 1,
        "key": "",
        "metric": "places_manager_write_query_count",
        "metric_type": "counter",
        "os": "*",
        "ping_type": "*",
        "range_max": 3.0,
        "range_min": 0.0,
    }
]

EXPECT = [
    {
        "agg_type": "histogram",
        "aggregates": [
            {"key": "1.00", "value": 0.166_666_666_666_666_66},
            {"key": "2.00", "value": 0.666_666_666_666_666_6},
            {"key": "4.00", "value": 0.166_666_666_666_666_66},
        ],
        "app_build_id": "*",
        "app_version": 84,
        "channel": "*",
        "client_agg_type": "count",
        "key": "",
        "metric": "places_manager_write_query_count",
        "metric_type": "counter",
        "os": "*",
        "ping_type": "*",
        "total_users": 1,
    }
]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__scalar_bucket_counts_v1.yaml",
        SCALAR_BUCKET_COUNTS,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)



"""Import Stripe reports into BigQuery."""

import os
import sys
import warnings
from datetime import datetime, timezone
from pathlib import Path
from tempfile import TemporaryFile
from time import sleep
from typing import Any, Dict, List, Optional

import requests
import rich_click as click
import stripe
import ujson
from dateutil.relativedelta import relativedelta
from google.cloud import bigquery
from requests.auth import HTTPBasicAuth


def _get_report_rows(
    api_key: Optional[str],
    after_date: Optional[datetime],
    before_date: Optional[datetime],
    report_type: str,
    columns: List[str],
):
    if api_key is None:
        yield from sys.stdin.buffer
    else:
        stripe.api_key = api_key
        parameters: Dict[str, Any] = {"columns": columns}
        if after_date:
            parameters["interval_start"] = int(after_date.timestamp())
        if before_date:
            parameters["interval_end"] = int(before_date.timestamp())

        try:
            run = stripe.reporting.ReportRun.create(
                report_type=report_type,
                parameters=parameters,
            )
        except stripe.error.InvalidRequestError as e:
            # Wrap exception to hide unnecessary traceback
            raise click.ClickException(str(e))

        click.echo(f"Waiting on report {run.id!r}", file=sys.stderr)
        # wait up to 30 minutes for report to finish
        timeout = datetime.utcnow() + relativedelta(minutes=30)
        while datetime.utcnow() < timeout:
            if run.status != "pending":
                break
            sleep(10)
            run.refresh()
        if run.status != "succeeded":
            raise click.ClickException(
                f"Report {run.id!r} did not succeed, status was {run.status!r}"
            )
        response = requests.get(
            run.result.url, auth=HTTPBasicAuth(api_key, ""), stream=True
        )
        response.raise_for_status()
        yield from (line + b"\n" for line in response.iter_lines())


@click.group("stripe", help="Commands for Stripe ETL.")
def stripe_():
    """Create the CLI group for stripe commands."""
    pass


@stripe_.command("import", help=__doc__)
@click.option(
    "--api-key",
    help="Stripe API key to use for authentication; If not set resources will be read "
    "from stdin",
)
@click.option(
    "--date",
    type=click.DateTime(formats=["%Y-%m-%d", "%Y-%m", "%Y"]),
    help="Creation date of resources to pull from stripe API; Added to --table "
    "to ensure only that date partition is replaced",
)
@click.option(
    "--after-date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Only pull resources from stripe API with a creation date on or after this; "
    "Used when importing resources older than the earliest available events",
)
@click.option(
    "--before-date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Only pull resources from stripe API with a creation date before this; "
    "Used when importing resources older than the earliest available events",
)
@click.option(
    "--table",
    help="BigQuery Standard SQL format table ID where resources will be written; "
    "if not set resources will be written to stdout",
)
@click.option(
    "--quiet",
    is_flag=True,
    help="Write output to os.devnull instead of sys.stdout",
)
@click.option(
    "--report-type",
    help="Stripe report type to import",
    required=True,
)
@click.option(
    "--time-partitioning-field",
    default="created",
    help="Field to use for partitioning and clustering; if --date or --before-date or"
    " --after-date are specified, values must fall within that window",
)
@click.option(
    "--time-partitioning-type",
    default=bigquery.TimePartitioningType.DAY,
    type=click.Choice(
        [
            bigquery.TimePartitioningType.DAY,
            bigquery.TimePartitioningType.MONTH,
            bigquery.TimePartitioningType.YEAR,
        ]
    ),
    help="BigQuery time partitioning type for --table",
)
def stripe_import(
    api_key: Optional[str],
    date: Optional[datetime],
    after_date: Optional[datetime],
    before_date: Optional[datetime],
    table: Optional[str],
    quiet: bool,
    report_type: str,
    time_partitioning_field: str,
    time_partitioning_type: str,
):
    """Import Stripe data into BigQuery."""
    if after_date:
        after_date = after_date.replace(tzinfo=timezone.utc)
    if before_date:
        before_date = before_date.replace(tzinfo=timezone.utc)
    if date:
        date = date.replace(tzinfo=timezone.utc)
        if time_partitioning_type == bigquery.TimePartitioningType.DAY:
            after_date = date
            before_date = after_date + relativedelta(days=1)
            if table:
                table = f"{table}${date:%Y%m%d}"
        elif time_partitioning_type == bigquery.TimePartitioningType.MONTH:
            after_date = date.replace(day=1)
            before_date = after_date + relativedelta(months=1)
            if table:
                table = f"{table}${date:%Y%m}"
        elif time_partitioning_type == bigquery.TimePartitioningType.YEAR:
            after_date = date.replace(month=1, day=1)
            before_date = after_date + relativedelta(years=1)
            if table:
                table = f"{table}${date:%Y}"

    if table:
        handle = TemporaryFile(mode="w+b")
    elif quiet:
        handle = open(os.devnull, "w+b")
    else:
        handle = sys.stdout.buffer  # type: ignore
    with handle as file_obj:
        path = Path(__file__).parent / f"{report_type}.schema.json"
        root = bigquery.SchemaField.from_api_repr(
            {"name": "root", "type": "RECORD", "fields": ujson.loads(path.read_text())}
        )
        columns = [f.name for f in root.fields]
        for row in _get_report_rows(
            api_key, after_date, before_date, report_type, columns
        ):
            file_obj.write(row)
        if table:
            if file_obj.writable():
                file_obj.seek(0)
            warnings.filterwarnings("ignore", module="google.auth._default")
            job_config = bigquery.LoadJobConfig(
                clustering_fields=[time_partitioning_field],
                ignore_unknown_values=False,
                time_partitioning=bigquery.TimePartitioning(
                    field=time_partitioning_field, type_=time_partitioning_type
                ),
                write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
                source_format=bigquery.SourceFormat.CSV,
                skip_leading_rows=1,
                schema=root.fields,
            )
            if "$" in table:
                job_config.schema_update_options = [
                    bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION
                ]
            job = bigquery.Client().load_table_from_file(
                file_obj=file_obj,
                destination=table,
                job_config=job_config,
            )
            try:
                click.echo(f"Waiting for {job.job_id}", file=sys.stderr)
                job.result()
            except Exception as e:
                full_message = f"{job.job_id} failed: {e}"
                for error in job.errors or ():
                    message = error.get("message")
                    if message and message != getattr(e, "message", None):
                        full_message += "\n" + message
                raise click.ClickException(full_message)
            else:
                click.echo(f"{job.job_id} succeeded", file=sys.stderr)
